// 
// The Inspirer - An Arduino-based gadget that delivers inspirational qwuotes & music 
// Copyright (C) 2021 Michael Wessel 

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.
//

#include "SerialMP3Player.h"
#include <Wire.h>
#include "Adafruit_LEDBackpack.h"
#include <SdFat.h>
#include <SPI.h>
#include <WString.h> 

//
// Buttons 
//

#define NEXT_SONG_NEXT_CAT_BUTTON 14 
#define PREV_SONG_PREV_CAT_BUTTON 15
#define RND_CAT_VOL_UP_BUTTON 16
#define FIX_CAT_VOL_DOWN_BUTTON 17
#define SHIFT_BUTTON 2 

#define RANDOM_ANALOG_PIN A0 // 23 = PC0 

//
// SD Card
//

SdFat SD;
SdFile root;
SdFile cat_file;
SdFile quote_file; 

//
// Quotes
//

#define QMAXLENGTH 70
#define AMAXLENGTH 32
#define CMAXLENGTH 16 

char quote[QMAXLENGTH]; 
char author[AMAXLENGTH]; 
char cat[CMAXLENGTH]; 

char fcat[13]; 
byte maxcat = 0; 
byte catn = 0; // 0 = random 
byte curcatn = 0; // 0 = random 

int maxquote = 0; 
int quoten = 0; 

//
// MP3 player
//

#define TX 0
#define RX 1
#define SDCARD_CHIP_SELECT 10 // PB2

SerialMP3Player mp3(RX, TX);

boolean volChanged = false;

//
// Display Management
//

int vol = 17;
int on_led_pos = 0; 
boolean on_led_left = true; 

//
//
// 

#define DISP_SCROLL_SPEED 130
#define DISP_WAIT 500

Adafruit_AlphaNum4 disp1 = Adafruit_AlphaNum4();
Adafruit_AlphaNum4 disp2 = Adafruit_AlphaNum4();
Adafruit_AlphaNum4 disp3 = Adafruit_AlphaNum4();
Adafruit_AlphaNum4 disp4 = Adafruit_AlphaNum4();

//
// state logic for detecting master button + vol+ / vol- key like a shift key
// 

#define BUTTON_PRESS_TIME 700 

unsigned long masterButtonPressedTime = 0; 

//
//
//

unsigned long lastTime ;
unsigned long lastDispTime; 

char displaybuffer[16] = {' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '};

byte cur_string_index = 0; 
boolean disp_active = false; 
boolean string_shown = false; 

char* cur_string = (char*) 0; 
char* next_string = (char*) 0; 
char* next_next_string = (char*) 0; 

//
//
//

void clear_display_buffer() {
  for (int i = 15; i >= 0; i--) {
    displaybuffer[i] = 32;  
  }
}

void write_displays() {
  disp1.writeDisplay(); 
  disp2.writeDisplay(); 
  disp3.writeDisplay(); 
  disp4.writeDisplay(); 
}

void show_display_buffer() {
  disp4.writeDigitAscii(0, displaybuffer[15]); 
  disp4.writeDigitAscii(1, displaybuffer[14]); 
  disp4.writeDigitAscii(2, displaybuffer[13]); 
  disp4.writeDigitAscii(3, displaybuffer[12]); 

  disp3.writeDigitAscii(0, displaybuffer[11]); 
  disp3.writeDigitAscii(1, displaybuffer[10]); 
  disp3.writeDigitAscii(2, displaybuffer[9]); 
  disp3.writeDigitAscii(3, displaybuffer[8]); 

  disp2.writeDigitAscii(0, displaybuffer[7]); 
  disp2.writeDigitAscii(1, displaybuffer[6]); 
  disp2.writeDigitAscii(2, displaybuffer[5]); 
  disp2.writeDigitAscii(3, displaybuffer[4]); 

  disp1.writeDigitAscii(0, displaybuffer[3]); 
  disp1.writeDigitAscii(1, displaybuffer[2]); 
  disp1.writeDigitAscii(2, displaybuffer[1]); 
  disp1.writeDigitAscii(3, displaybuffer[0]); 

  write_displays(); 
} 

void queue_char(char c) {

  for (int i = 15; i > 0; i--) {
    displaybuffer[i] = displaybuffer[i-1]; 
  }
  displaybuffer[0] = c; 
  show_display_buffer(); 
}

void display_string(String msg) {
  clear_display_buffer();   
  byte i = 0; 
  while ( msg[i] ) {
    queue_char(msg[i]); 
    i++;     
  } 
}

void display_string0(String msg) {
  clear_display_buffer();   
  byte i = 0; 
  while ( msg[i] ) {
    displaybuffer[15-i] = msg[i]; 
    i++; 
  }
}

void ascii_at(int pos, char c) {
  displaybuffer[pos] = c; 
}

void display_disp(String msg, Adafruit_AlphaNum4 disp) {

  disp.writeDigitAscii(0, msg[0]); 
  disp.writeDigitAscii(1, msg[1]); 
  disp.writeDigitAscii(2, msg[2]); 
  disp.writeDigitAscii(3, msg[3]); 

  disp.writeDisplay(); 
 
}

void clear_all() {

  clear_display_buffer(); 
  
  disp1.clear();
  disp2.clear();
  disp3.clear();
  disp4.clear();
  
  write_displays(); 

}

//
//
//

void queue_disp_string(char* msg) {
  cur_string = msg; 
  cur_string_index = 0; 
  string_shown = false; 
  lastDispTime = millis(); 
  lastTime = lastDispTime; 
  disp_active = true; 
}


//
//
//

int decodeHex(char c) {
  if (c >= 65 && c <= 70 )
    return c - 65 + 10;
  else if ( c >= 48 && c <= 67 )
    return c - 48;
  else return -1;
}

int catfile_readInt() {   
  int n = 0; 
  while (true) {
    int x = cat_file.read(); 
    if (x == -1)
      return n; 
    if ( x == 0x0d) { // CR
      cat_file.read(); // LF 
      return n;
    } else {
      n = n*10 + decodeHex(x); 
    }
  }        
}

void catfile_readString(char* buffer) {
  int n = 0; 
  while (true) {
    char x = cat_file.read(); 
    if (x == -1)
      return; 
    if ( x == 0x0d) { // CR
      cat_file.read(); // LF 
      buffer[n] = 0; // end of string
      return; 
    } else {
      buffer[n++] = x; 
    }
  }      
}



int quotefile_readInt() {   
  int n = 0; 
  while (true) {
    int x = quote_file.read(); 
    if (x == -1)
      return n; 
    if ( x == 0x0d) { // CR
      quote_file.read(); // LF 
      return n;
    } else {
      n = n*10 + decodeHex(x); 
    }
  }        
}

void quotefile_readString(char* buffer) {
  int n = 0; 
  while (true) {
    char x = quote_file.read(); 
    if (x == -1)
      return; 
    if ( x == 0x0d) { // CR
      quote_file.read(); // LF 
      buffer[n] = 0; // end of string
      return; 
    } else {
      if (isprint(x))
        buffer[n++] = x; 
    }
  }      
}


void read_category() {

  SD.vwd()->rewind();

  cat_file.open( "mapping.txt", FILE_READ);
  if ( cat_file.isOpen()) {

    maxcat = catfile_readInt(); 

    // random? 
    curcatn = catn ? catn : random(maxcat) + 1;     
      
    for (int i = 1; i <= curcatn; i++) {
      catfile_readString(cat);    
      catfile_readString(fcat);      
    }   

    int j = 0; 
    while (fcat[j]) {
      j++;
    } 

    fcat[j++] = '.';
    fcat[j++] = 't';
    fcat[j++] = 'x';
    fcat[j++] = 't';
    fcat[j] = 0; 

    cat_file.close(); 

  }
}


void read_quote(int quoten) {
  for (int i = 1; i <= quoten; i++) {
    quotefile_readString(quote); 
    quotefile_readString(author);    
  }
}

void get_no_of_categories() {
  SD.vwd()->rewind();
  cat_file.open( "mapping.txt", FILE_READ);
  if ( cat_file.isOpen()) {
    maxcat = catfile_readInt(); 
    cat_file.close(); 
  }
}


void random_quote() {
  SD.vwd()->rewind();

  quote_file.open( fcat, FILE_READ);

  if ( quote_file.isOpen()) {
    
    maxquote = quotefile_readInt(); 
    quoten = random(maxquote) + 1; 

    read_quote(quoten); 
    quote_file.close();

  }
}

void trigger_random_quote() {
      
  read_category(); 
  delay(100); 
  random_quote(); 
  queue_disp_string(cat);     
  next_string = quote; 
  next_next_string = author; 

}

//
// Hardware Initialization
//

void setup() {
  
  randomSeed(analogRead(RANDOM_ANALOG_PIN));

  // after that, we change A0 = PC0 to a digital input: 

  pinMode(NEXT_SONG_NEXT_CAT_BUTTON, INPUT_PULLUP); // PC0 = Play Next Song / Next Category
  pinMode(PREV_SONG_PREV_CAT_BUTTON, INPUT_PULLUP); // PC1 = Play Prev Song / Prev Category
  pinMode(RND_CAT_VOL_UP_BUTTON, INPUT_PULLUP); // PC2 = Random   Category / Increase Volume 
  pinMode(FIX_CAT_VOL_DOWN_BUTTON, INPUT_PULLUP); // PC3 = Fix Cur. Category / Decrease Volume 
  pinMode(SHIFT_BUTTON, INPUT_PULLUP);  // PD1 = Shift Button 

  //
  // MP3 init 
  // 

  mp3.begin(9600);        // start mp3-communication
  delay(500);             // wait for init

  mp3.sendCommand(CMD_SEL_DEV, 0, 2);   //select sd-card
  delay(500);             // wait for init

  //
  // Displays init 
  // 

  disp1.begin(0x74);  // pass in the address
  disp2.begin(0x71);  // pass in the address
  disp3.begin(0x70);  // pass in the address
  disp4.begin(0x72);  // pass in the address

  clear_all();  

  //
  // SDcard init 
  // 
 
  if (! SD.begin(SDCARD_CHIP_SELECT)) {
    display_string(F("-FIX BAD SDCARD-"));      
    delay(DISP_WAIT); 
         
  } else {
    display_string(F("*Inspirer Ready*"));      
    delay(DISP_WAIT); 
    get_no_of_categories(); 
  }

  //
  // Set initial MP3 volume 
  //

  mp3.setVol(vol);
  mp3.qVol();

  //
  // Good to go! 
  // 

  lastTime = millis();
  lastDispTime = millis(); 
  clear_all(); 

}

//
// Main Loop 
//

void loop() {

  if ((millis() - lastTime) > DISP_WAIT) {
     
    if (volChanged) {
      masterButtonPressedTime = 0; 

      mp3.setVol(vol);
      mp3.qVol();

      display_string0(F("Volume is now "));      
      ascii_at(1, 48 + vol / 10);
      ascii_at(0, 48 + vol % 10);
      show_display_buffer();          

      volChanged = false;

    } else {

      //
      // show activity led cylon eye 
      // 

      if (! disp_active && ! volChanged) {

	clear_display_buffer();       
	ascii_at(on_led_pos, '*'); 
	if (on_led_left) {
	  if (on_led_pos == 15){
	    on_led_left = false; 
	    on_led_pos--; 
	    trigger_random_quote(); 
	  } else 
	    on_led_pos++; 
	} else {
	  if (on_led_pos == 0){
	    on_led_left = true; 
	    on_led_pos++; 
	    trigger_random_quote(); 
	  } else 
	    on_led_pos--; 
	}
	show_display_buffer(); 
      }		   
    }

    //
    //
    // 

    lastTime = millis();

  }

  //
  //
  //

  if (disp_active) {
    if ((millis() - lastDispTime) > DISP_SCROLL_SPEED) {      
      if ( ! string_shown && cur_string[cur_string_index] ) {
        queue_char(cur_string[cur_string_index++]);     
      } else {       

        if (! string_shown) 
          cur_string_index = 0; 
        else
          cur_string_index++;        
	  
        string_shown = true;      

        if (cur_string_index  == 16) {
          disp_active = false; 
          if ( next_string[0] ) {            
            queue_disp_string(next_string); 
            next_string = next_next_string;                                    
            next_next_string = (char*) 0;             
          }        
        } else 
          queue_char(' ');  
      }      
      lastDispTime = millis(); 
    }
  }

  //
  //
  //

  if (! digitalRead(NEXT_SONG_NEXT_CAT_BUTTON)) {
    if (! digitalRead(SHIFT_BUTTON)) {
      masterButtonPressedTime = 0; 
      if (catn < maxcat)
	catn++; 
      else 
	catn = 1; 
      read_category(); 
      display_string(cat);      
    } else {
      display_string(F("Play Next Song >")); 
      mp3.playNext();
    }
    delay(DISP_WAIT); 
    lastTime = millis();
  } else if (! digitalRead(PREV_SONG_PREV_CAT_BUTTON)) {
    if (! digitalRead(SHIFT_BUTTON)) {
      masterButtonPressedTime = 0; 
      if (catn > 1)
	catn--; 
      else 
	catn = maxcat; 
      read_category(); 
      display_string(cat); 
    } else {
      display_string(F("Play Prev Song <")); 
      mp3.playPrevious();
    }
    delay(DISP_WAIT); 
    lastTime = millis();
  } else if (! digitalRead(RND_CAT_VOL_UP_BUTTON)) {
    if (! digitalRead(SHIFT_BUTTON)) {
      volChanged = true;
      masterButtonPressedTime = 0; 
      if (vol < 41) 
	vol++;
      display_string0(F("Change Volume "));      
      ascii_at(1, 48 + vol / 10);
      ascii_at(0, 48 + vol % 10);
      show_display_buffer();          
    } else {      
      catn = 0; 
      read_category(); 
       display_string(F("Random Category ")); 
    }
    delay(DISP_WAIT); 
    lastTime = millis();
  } else if (! digitalRead(FIX_CAT_VOL_DOWN_BUTTON) ) {
    if (! digitalRead(SHIFT_BUTTON)) {
      volChanged = true;
      masterButtonPressedTime = 0; 
      if (vol > 0) 
	vol--;
      display_string0(F("Change Volume ")); 
      ascii_at(1, 48 + vol / 10);
      ascii_at(0, 48 + vol % 10);
      show_display_buffer();       
    } else {      
      catn = curcatn; 
      read_category(); 
       display_string(F("Fix Category    ")); 
    } 
    delay(DISP_WAIT); 
    lastTime = millis();

  } else if (! digitalRead(SHIFT_BUTTON)) {

    if ( ! masterButtonPressedTime )
      masterButtonPressedTime = millis(); 

    else if ((millis() - masterButtonPressedTime) > BUTTON_PRESS_TIME) {
      masterButtonPressedTime = 0; 
      trigger_random_quote(); 
    }
  } else {
    masterButtonPressedTime = 0; 
  }


}
